var Resources = function (runtime) {
    var bridge = runtime.bridge;
    var module = runtime.module;

    var Model = {
        create: function (name) {
            return {
                id: -1,
                name: name
            };
        }
    };

    var Texture = {
        create: function (name) {
            var self = {
                id: bridge.GetNextTextureId(),
                name: name
            };

            return self;
        }
    }

    var Material = {
        create: function (path, name) {
            return {
                id: bridge.LoadMaterial(path),
                name: name
            };
        }
    }

    return {
        runtime: runtime,
        createMesh: function () { return Mesh.create(bridge); },
        loadMaterial: function (path, name) { return Material.create(path, name); },
        loadMaterialInstance: function (material, name) { return MaterialInstance.create(bridge, material, name); },
        loadModel: function (path) {
            var model = Model.create();
            model.path = path;

            model.id = runtime.bridge.LoadModel(model);

            return model;
        },
        loadTexture: function (bytes, width, height, format, stride, flags) {
            var texture = Texture.create();

            runtime.bridge.LoadTexture(texture.id, bytes, width, height, format, stride, flags);

            return texture;
        },
        createColorTexture: function (bytes) {
            return this.loadTexture(bytes, bytes.length / 4, 1, TextureFormat.R8G8B8A8_UNORM, bytes.length, TextureLoadFlags.None);
        },
        loadTextureImage: function (image) {
            var texture = Texture.createExternal();
            loadExternalTextureFromImg(texture, image);
            return texture;
        },
        loadTextureUrl: function (url) {
            var texture = Texture.create();

            runtime.bridge.LoadTextureResource(texture.id, url, TextureLoadFlags.None);

            return texture;
        },
    };
};
