
var Runtime = function () {
    var runtime = {
        entities: {},
        entitiesById: {},
        clear: function () {
            runtime.bridge.ClearContext();
            runtime.entities = {};
            runtime.entitiesById = {};
            runtime.bridge.SetupContext();
        },
        destroy: function (entity) {
            runtime.bridge.DestroyEntity(entity.id);

            if (entity.name) {
                runtime.entities[entity.name] = undefined;
            }

            runtime.entitiesById[entity.id] = undefined;
        },
        loadScene: function (path) {
            bridge.LoadScene(path, 1);
        },
        setVariable: function (key, value) {
            bridge.SetVariable(key, value);
        },
        getVariable: function (key) {
            return bridge.GetVariable(key);
        },
        initializeTypes: function () {
            var bridge = runtime.bridge;

            runtime.SetterTypeMap = {
                "bool": bridge.SetBoolField,
                "string": bridge.SetStringField,
                "float": bridge.SetFloatField,
                "double": bridge.SetDoubleField,
                "int32_t": bridge.SetIntegerField,
                "uint32_t": bridge.SetIntegerField,
                "vec2": bridge.SetVector2Field,
                "vec3": bridge.SetVector3Field,
                "vec4": bridge.SetVector4Field,
                "quat": bridge.SetQuaternionField,
                "mat4": bridge.SetMatrixField,
                "dvec2": bridge.SetVector2dField,
                "dvec3": bridge.SetVector3dField,
                "dvec4": bridge.SetVector4dField,
                "dquat": bridge.SetQuaterniondField,
                "uvec3": bridge.SetVector3Field,
                "vector<int32_t>": bridge.SetMultiIntegerField,
                "vector<float>": bridge.SetMultiFloatField,
                "vector<vec2>": bridge.SetMultiVector2Field,
                "vector<vec3>": bridge.SetMultiVector3Field,
                "vector<vec4>": bridge.SetMultiVector4Field,
                "vector<string>": bridge.SetMultiStringField,
                "shared_entity": bridge.SetEntityField,
                "weak_shared_entity": bridge.SetEntityField,
                "vector<shared_entity>": bridge.SetChildren,
                "vector<weak_shared_entity>": bridge.SetMultiEntityField,
                // Resources
                "TextureHandle": bridge.SetTextureField,
                "MeshHandle": bridge.SetMeshField,
                "ModelHandle": bridge.SetModelField,
                "FontHandle": bridge.SetFontField,
                "BufferResourceHandle": bridge.SetBufferResourceField,
                "GuiDocumentHandle": bridge.SetGuiDocumentField,
                "MaterialInstanceHandle": bridge.SetMaterialInstanceField,
                // Enums
                "CullMode": bridge.SetIntegerField,
                "AddressMode": bridge.SetIntegerField,
                "RenderLayers": bridge.SetIntegerField,
                "ProjectionMode": bridge.SetIntegerField,
                "ShapeType": bridge.SetIntegerField,
                "LodPolicy": bridge.SetIntegerField,
                "PrimitiveType": bridge.SetIntegerField,
                "PositionMode": bridge.SetIntegerField,
                "VerticalAlignment": bridge.SetIntegerField,
                "ShapePrimitiveType": bridge.SetIntegerField,
                "LightType": bridge.SetIntegerField,
                // Casing extension
                //"Parameterization": bridge.SetIntegerField,
                //"LoftShape": bridge.SetIntegerField,
                "ImageType": bridge.SetIntegerField,
                //"GuiPointerSource": bridge.SetIntegerField,
                "CameraFlags": bridge.SetIntegerField,
                "LightingLayers": bridge.SetIntegerField,
            };

            var createFieldGetter = function (arrayType, size) {
                return function (entity, component, field, value) { return bridge.GetFieldValue(entity, component, field, value, arrayType, size); }
            }

            runtime.GetterTypeMap = {
                "bool": bridge.GetBoolFieldValue,
                "uint32_t": createFieldGetter(Uint32Array, 1),
                "float": createFieldGetter(Float32Array, 1),
                "vec2": createFieldGetter(Float32Array, 2),
                "vec3": createFieldGetter(Float32Array, 3),
                "vec4": createFieldGetter(Float32Array, 4),
                "mat4": createFieldGetter(Float32Array, 16),
                "array<MaterialInstanceHandle>": bridge.GetMaterialInstances,
                "MaterialInstanceHandle": bridge.GetMaterialInstance,
                "vector<shared_entity>": bridge.GetChildren,
            };

            var intGetter = createFieldGetter(Int32Array, 1);
            for (var fieldType in runtime.SetterTypeMap) {
                if (runtime.SetterTypeMap[fieldType] == bridge.SetIntegerField) {
                    runtime.GetterTypeMap[fieldType] = intGetter;
                }
            }

            runtime.IdMap = {};
            runtime.GetterIdMap = {};

            for (var fieldType in runtime.SetterTypeMap) {
                var typeId = runtime.bridge.getTypeId(fieldType);
                runtime.IdMap[typeId] = runtime.SetterTypeMap[fieldType];

                if (runtime.GetterTypeMap.hasOwnProperty(fieldType)) {
                    runtime.GetterIdMap[typeId] = runtime.GetterTypeMap[fieldType];
                }
            }
        },
        initialize: function (onEngineInitialized, settings) {
            runtime.bridge = bridge;

            try {
                runtime.initializeTypes(runtime);

                runtime.entityHandler = EntityHandler;
                runtime.entityHandler.initializeEntities(runtime);

                runtime.componentHandler = ComponentHandler;
                runtime.componentHandler.initializeComponents(runtime, settings.extensions);

                //runtime.scene = Scene(runtime);
                runtime.resources = Resources(runtime);
                runtime.input = Input.create(runtime, runtime.bridge);

                if (onEngineInitialized) onEngineInitialized();
            } catch (e) {
                console.error(e.stack);
            }
        },
        initializeModule: function (canvas, settings, init) {

            this.module = {};

            return this.module;
        }
    };

    console.log("Created JS runtime.");

    return runtime;
}
