#include "AdaptivePlanarGrid.es20.glsl"

#ifdef COGS_BASIC_OCEAN_QUUX
#include "BasicOceanEffectQuux.es20.glsl"
#else
#include "BasicOceanEffect.es20.glsl"
#endif

VertexOut VertexFunction(VertexIn In)
{
  // Snap edge vertices to neighbouring grids
  vec2 pos = snapAdaptivePlanarGridEdges(In.Position.xyz).xy;

  // Move and scale patch to its position in world space
  vec3 world = (worldMatrix * vec4(pos, 0, 1)).xyz;

  // Calculate coordinates in wave space.
  vec2 waveCoordinates = getTexCoordWorld(pos.xy);

  // And we only displace if we are relatively close to the camera
  vec3 surfacePositionWorld = world;
  vec4 surfacePositionView = viewMatrix * vec4(surfacePositionWorld, 1.0);

  // Determine pixel size of significant waveheight, which we use to control LoD
#ifdef COGS_BASIC_OCEAN_QUUX
  vec4 a0 = worldToClipMatrix * vec4(world, 1);
  vec4 a1 = worldToClipMatrix * vec4(world + significantWaveHeight() * cameraYAxis().xyz, 1);
  vec2 h = viewportScale() * ((1.0 / a0.w)*a0.xy - (1.0 / a1.w)*a1.xy);
  float lod = length(h);

  surfacePositionWorld += waveDisplace(world.xy, lod);

#else
  vec4 a0 = worldToClipMatrix * vec4(world, 1.0);
  vec4 a1 = worldToClipMatrix * vec4(world + cameraYAxis().xyz, 1.0);
  float lod = significantWaveHeight() * cameraYAxis().w * ((a1.y / a1.w) - (a0.y / a0.w));
  if (lod > 2.0) {
    float blend = clamp(lod - 2.0, 0.0, 1.0);
    surfacePositionWorld += blend * waveDisplace(waveCoordinates);
  }
#endif

  surfacePositionWorld.z += seaLevel();


  // Output results
  VertexOut Out = createVertexOut();
  Out.undisplacedPositionWorld = world.xy;

#ifdef COGS_BASIC_OCEAN_CURVED
  // Assume TransformSystem's origin is tangent on earth
  float R = 6371000.0;
  //R *= 1.0 / 200;
  float d = length(world.xy);
  surfacePositionWorld.z -= R - sqrt((R - d)*(R + d));
#endif


  Out.Position = worldToClipMatrix * vec4(surfacePositionWorld, 1.0);
  Out.WorldPosition = vec4(surfacePositionWorld, 1);

  Out.screenPosition = Out.Position;
  Out.wavePosition = vec3(waveCoordinates, lod);

  return Out;
}
