float calculateStippleAlpha(float dist)
{
  float fi;
  float f = modf(dist, fi);
  uint i = fi;
  i /= stippleFactor * (uint)lineWidth;

  return (stipplePattern & (1 << i)) != 0 ? 1.0 : 0.0;
}

float calculateEndpointAlpha(float distance, float length)
{
  float margin = lineFeather / (2 * lineFeather + length);

  if (distance < 0.5) {
    return saturate(distance / margin);
  } else {
    return saturate((length - distance) / margin);
  }
}

float calculateSideAlpha(float weight)
{
  float margin = lineFeather / (2 * lineFeather + lineWidth);

  if (weight < 0.5) {
    return saturate(weight / margin);
  } else {
    return saturate((1 - weight) / margin);
  }
}

SurfaceOut surfaceFunction(VertexIn In)
{
  SurfaceOut Out = (SurfaceOut)0;

  float4 color = diffuseColor;

#ifdef COGS_LINE_TEXTURE_ENABLE
  color *= diffuseMap.Sample(diffuseMapSampler, In.TexCoord.xy);
#endif

#ifdef COGS_DEFAULT_MATERIAL_TEXTURED
  color *= diffuseMap.Sample(diffuseMapSampler, In.TexCoord.xy);
#endif
  
#ifdef DEFAULT_MATERIAL_VERTEX_COLOR
  color.rgb *= In.Color.rgb;
#endif

  float a = color.a;

#ifndef COGS_MSAA
  a *= calculateSideAlpha(In.Weights.x);
  a *= In.Weights.y < 0.001 ? calculateEndpointAlpha(In.Weights.z, In.Weights.w) : 1;
#endif

  a *= calculateStippleAlpha(In.Weights.z);

  if (a <= 0) discard;

  Out.Color = float4(color.rgb, a);
  Out.Emissive = emissiveColor.rgb;
  Out.Normal = float3(1, 1, 1);

  return Out;
}
