#include "Bayer.es20.glsl"

#define CASCADE_SHADOW_LINE               0
#define CASCADE_SHADOW_NEAR_SPLITS        1
#define CASCADE_SHADOW_FAR_SPLITS         2
#define CASCADE_SHADOW_INDEX_BIAS_CAS0TXY 3
#define CASCADE_SHADOW_CAS3_ROW0          4
#define CASCADE_SHADOW_CAS3_ROW1          5
#define CASCADE_SHADOW_CAS3_ROW2          6
#define CASCADE_SHADOW_CAS0SXYZ_CAS0TZ    7
#define CASCADE_SHADOW_CAS1SXYZ_CAS1TZ    8
#define CASCADE_SHADOW_CAS2SXYZ_CAS2TZ    9
#define CASCADE_SHADOW_CAS1TXY_CAS2TXY    10
#define CASCADE_SHADOW_DATA_N             11

#define POINT_SHADOW_NEAR_FAR             0

uniform vec4 shadowData[CASCADE_SHADOW_DATA_N];

uniform sampler2D cascadedShadowMap;

float cascadeContribution(in vec4 splits, in float z, in vec4 screenPos)
{
  return splits.y <= z && z < splits.z ? 1.0 : 0.0;
}

float shadowSample(in vec2 tileOff, in float tileSize, in vec2 unitPos, in float z)
{
  float shadowness = 0.0;
  if (all(lessThanEqual(vec2(0.0), unitPos)) && all(lessThanEqual(unitPos, vec2(1.0)))){
    vec2 tiledPos = tileSize * (unitPos + tileOff);
    shadowness = texture2D(cascadedShadowMap, tiledPos).r < z ? 1.0 : 0.0;
  }
  return shadowness;
}

float shadowSampleBilinear(in vec2 tileOff, in float tileSize, in vec2 unitPos, in float z, in float texelSize)
{
  vec2 pixelPos = floor(unitPos / texelSize);
  vec2 b = unitPos / texelSize - pixelPos;

  return
    shadowSample(tileOff, tileSize, texelSize*(pixelPos + vec2(0, 0)), z)*(1.0 - b.x)*(1.0 - b.y) +
    shadowSample(tileOff, tileSize, texelSize*(pixelPos + vec2(1, 0)), z)*b.x*(1.0 - b.y) +
    shadowSample(tileOff, tileSize, texelSize*(pixelPos + vec2(0, 1)), z)*(1.0 - b.x)*b.y +
    shadowSample(tileOff, tileSize, texelSize*(pixelPos + vec2(1, 1)), z)*b.x*b.y;
}

float calculateShadowContribution(in vec4 worldPosition, in vec3 normal, in vec3 lightDir, in int lightIndex, in vec4 screenPos)
{
#if defined(COGS_DISABLE_DIRECTIONAL_SHADOWS)
  return 1.0;
#endif

  if (int(shadowData[CASCADE_SHADOW_INDEX_BIAS_CAS0TXY].x) != lightIndex) return 1.0;

  int tileBaseIndex = 0;
  float sampleBias = shadowData[CASCADE_SHADOW_INDEX_BIAS_CAS0TXY].y;
  float tilesInRow = 2.0;
  float tileSize = 0.5;
  float texelSize = fract(shadowData[CASCADE_SHADOW_INDEX_BIAS_CAS0TXY].x);

  vec4 cascadeLine = shadowData[CASCADE_SHADOW_LINE];
  vec4 nearSplits = shadowData[CASCADE_SHADOW_NEAR_SPLITS];
  vec4 farSplits = shadowData[CASCADE_SHADOW_FAR_SPLITS];

  float viewZ = dot(cascadeLine.xyz, worldPosition.xyz) + cascadeLine.w;
  if (farSplits.w < viewZ) return 1.0;

  int cascadeIx = 3;
  float nearSplit;
  float farSplit;
  for (int c = 3; c>=0; --c) {
    if (viewZ < farSplits[c]) {
      cascadeIx = c;
      nearSplit = nearSplits[c];
      farSplit = farSplits[c];
    }
  }

  // FIXME: Use some blue-noise look-up instead.
  float fade = viewZ - nearSplit;
  if (0.0 < fade && cascadeIx < 3) {
    fade = (1.0 / (farSplit - nearSplit)) * fade;
    if (bayer4(screenPos.xy) < fade) {
      cascadeIx++;
    }
  }


  vec3 posLightSpace = vec3(dot(shadowData[CASCADE_SHADOW_CAS3_ROW0], vec4(worldPosition.xyz, 1)),
                            dot(shadowData[CASCADE_SHADOW_CAS3_ROW1], vec4(worldPosition.xyz, 1)),
                            dot(shadowData[CASCADE_SHADOW_CAS3_ROW2], vec4(worldPosition.xyz, 1)));
  if (cascadeIx == 0) {
    posLightSpace = shadowData[CASCADE_SHADOW_CAS0SXYZ_CAS0TZ].xyz * posLightSpace +
                    vec3(shadowData[CASCADE_SHADOW_INDEX_BIAS_CAS0TXY].zw,
                         shadowData[CASCADE_SHADOW_CAS0SXYZ_CAS0TZ].w);
  }
  else if (cascadeIx == 1) {
    posLightSpace = shadowData[CASCADE_SHADOW_CAS1SXYZ_CAS1TZ].xyz * posLightSpace +
                    vec3(shadowData[CASCADE_SHADOW_CAS1TXY_CAS2TXY].xy,
                         shadowData[CASCADE_SHADOW_CAS1SXYZ_CAS1TZ].w);
  }
  else if (cascadeIx == 2) {
    posLightSpace = shadowData[CASCADE_SHADOW_CAS2SXYZ_CAS2TZ].xyz * posLightSpace +
                    vec3(shadowData[CASCADE_SHADOW_CAS1TXY_CAS2TXY].zw,
                         shadowData[CASCADE_SHADOW_CAS2SXYZ_CAS2TZ].w);
  }

  posLightSpace.z -= sampleBias;

  vec2 tileOff = vec2(mod(float(cascadeIx), tilesInRow),
                      floor(float(cascadeIx) / tilesInRow));
#if defined(COGS_SOFT_SHADOWS_HIGH)
  float r = 1.5*texelSize;
  float g = 2.5*texelSize;
  float shadowness = (1.0 / 25.0)*(shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-g, -g), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-g, -r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-g, +0), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-g, +r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-g, +g), posLightSpace.z, texelSize) +

                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, -g), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, -r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, +0), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, +r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, +g), posLightSpace.z, texelSize) +

                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, -g), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, -r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, +0), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, +r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, +g), posLightSpace.z, texelSize) +

                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, -g), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, -r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, +0), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, +r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, +g), posLightSpace.z, texelSize) +

                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+g, -g), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+g, -r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+g, +0), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+g, +r), posLightSpace.z, texelSize) +
                                   shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+g, +g), posLightSpace.z, texelSize));
#elif defined(COGS_SOFT_SHADOWS_HIGH)
  float r = 1.5*texelSize;
  float shadowness = (1.0 / 9.0)*(shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, -r), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, +0), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(-r, +r), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, -r), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, +0), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+0, +r), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, -r), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, +0), posLightSpace.z, texelSize) +
                                  shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy + vec2(+r, +r), posLightSpace.z, texelSize));
#else
  float shadowness = shadowSampleBilinear(tileOff, tileSize, posLightSpace.xy, posLightSpace.z, texelSize);
#endif

  return max(0.0, 1.0 - shadowness);
}
