// uint3 : SV_GroupThreadID     Thread id within block
// uint3 : SV_GroupID           Block id
// uint3 : SV_DispatchThreadID  Global thread id
// uint  : SV_GroupIndex        Linearized thread id within block

// groupshared shmem[64];  Declaration of shared memory

//AllMemoryBarrier - block execution until all memory accesses has completed. (my writes visible to world?)
//AllMemoryBarrierWithGroupSync - Above + wait until threads in group have reached this call.
//DeviceMemoryBarrier
//DeviceMemoryBarrierWithGroupSync
//GroupMemoryBarrier
//GroupMemoryBarrierWithGroupSync();

cbuffer OffsetParameters
{
  uint NodeAlloc;
  uint width;
  uint height;
};

Texture2D<float> complexityTexture;

RWStructuredBuffer<uint>  countBuffer;
RWStructuredBuffer<uint>  offsetBuffer;

[numthreads(64,4,1)]
void main(uint3 gid : SV_GroupID, uint tix : SV_GroupIndex)
{
  // This atomic add stuff might be replaced with prefix-sum scan.

  uint ix = 64 * 4 * gid.x + tix;
  uint j = ix / width;
  if (j < height) {
    uint i = ix - width*j;

    uint complexity = complexityTexture.Load(int3(i, j, 0));

    uint offset;
    InterlockedAdd(countBuffer[0], complexity, offset);

    offsetBuffer[2 * ix + 0] = offset;
    offsetBuffer[2 * ix + 1] = offset + complexity;
  }
}