#include "Engine/TransparencyBlendCommon.hlsl"
#include "Engine/PhongLighting.hlsl"

Texture2D depthTexture;
SamplerState linearSampler;

ByteAddressBuffer               List;
StructuredBuffer<FragmentNode>  Node;
StructuredBuffer<FragmentData>  Data;

PixelOut main(VertexIn In, float4 ScreenPos : SV_Position)
{
  const float backgroundDepth = depthTexture.Load(int3(ScreenPos.xy, 0)).r;

  uint startOffset = 4 * ((uint)ScreenPos.y/HeightDivisor * Width + (uint)ScreenPos.x/WidthDivisor);
  uint rx = List.Load(startOffset);
  uint ix = rx & 0x7fffffffu;

  if (NodeAlloc <= ix) {
    PixelOut BackOut = { float4(0, 0, 0, 1) };
    return BackOut;
  }

  // Plow through the full fragment list for a pixel location _once_, while maintaining
  // a depth-sorted array of the MAX_SAMPLES nearest entries.
  uint indices[MAX_SAMPLES];
  float depths[MAX_SAMPLES];
  uint tailIx = 0;
  indices[0] = rx;
  depths[0] = Node[ix].Depth;
  rx = Node[ix].Next;
  ix = rx & 0x7fffffffu;
  uint maxIter = 100;
  uint currentIter = 0;

  while (ix < NodeAlloc) {
    // Driver crash safeguard.
    if (currentIter++ == maxIter) break;

    float dp = Node[ix].Depth;
    bool notFull = tailIx < MAX_SAMPLES - 1;
    bool closerThanTail = dp < depths[tailIx];
    if (notFull || closerThanTail) {
      tailIx = tailIx + (notFull ? 1 : 0);
      int j = tailIx;
      for (int i = tailIx - 1; i >= 0; i--) {
#ifdef COGS_REVERSE_DEPTH
        if (dp > depths[i]) {
#else
        if (dp < depths[i]) {
#endif
          indices[i + 1] = indices[i];
          depths[i + 1] = depths[i];
          j = i;
        }
      }
      indices[j] = rx;
      depths[j] = dp;
    }
    rx = Node[ix].Next;
    ix = rx & 0x7fffffffu;
  }

  float4 result = float4(0, 0, 0, 0);
  for (uint i = 0; (i <= tailIx) && (result.a < 0.99); i++) {
    uint ix = indices[i];

    float depth = depths[i];
    FragmentData data = Data[ix & 0x7fffffffu];

    float4 viewSpacePosition = mul(float4(In.ScreenPosition.xy, depth, 1), inverseProjectionMatrix);
    viewSpacePosition /= viewSpacePosition.w;

    SurfaceLighting lightSurface = (SurfaceLighting)0;
    lightSurface.ViewSpacePosition = viewSpacePosition.xyz;
    unpack(lightSurface.Normal,
           lightSurface.LightModel,
           lightSurface.Color.a,
           lightSurface.Emissive,
           lightSurface.Color.rgb,
           lightSurface.Specular,
           lightSurface.SpecularPower,
           data);

    float4 shaded = float4(DoLighting(lightSurface, ScreenPos).rgb, lightSurface.Color.a);

    float o = 1.0 - result.a;
    result.rgb = result.rgb + o * shaded.rgb;
    result.a = min(1, result.a + o * shaded.a);
  }

  PixelOut Out;
  Out.Color.rgb = result.rgb;
  Out.Color.a = (1.0-result.a);
  return Out;
}
