#include "Common.es20.glsl"

#include "Noise.es20.glsl"

uniform highp vec4 cameraYAxis; // x,y,z=camera y axis in world space, w=0.5*viewportsize.y
uniform highp vec4 waterColor;
uniform highp vec2 waveDirection;
uniform highp vec2 camPlaneDir;

uniform highp float significantWaveHeight;
uniform highp float fftTileScale;
uniform highp float camAzimuth;

uniform highp float phaseShiftNoiseFrequency;
uniform highp float phaseShiftNoisePeriod;

uniform sampler2D Displacement;
uniform sampler2D Difference;

uniform sampler2D Tangents;
uniform sampler2D PlanarReflection;

varying vec3 vertex_undisplacedPositionWorld;
varying vec3 vertex_displacedPositionWorld;
varying vec4 vertex_screenPosition;
varying vec2 vertex_wavePosition;

vec4 calculateWaveParameters(vec2 world)
{
  vec4 w;
  w.xy = fftTileScale * world;

  float theta = 2.1 * pnoise(vec3(phaseShiftNoiseFrequency * w.xy, 0.5), vec3(vec2(phaseShiftNoisePeriod), 1));
  w.zw = vec2(cos(theta), sin(theta));

  return w;
}

vec3 waveDisplace(in vec2 world)
{
  vec4 w = calculateWaveParameters(world);
  
  vec2 b = w.xy;
  vec2 phase = w.zw;

  vec3 real = texture2D(Displacement, b).rgb;
  vec3 imag = texture2D(Difference, b).rgb;

  float z = -(real.z*phase.x - imag.z*phase.y);
  vec2 xy = real.xy*phase.y + imag.xy*phase.x;

  return significantWaveHeight*vec3(xy, z);
}

vec3 waveNormal(in vec2 world)
{
  vec4 w = calculateWaveParameters(world);
  
  vec2 b = w.xy;
  vec2 phase = w.zw;

  vec4 tmp = significantWaveHeight*(texture2D(Tangents, b) +
                                    0.5*texture2D(Tangents, 16.0*b));

  vec2 rxy = (tmp.xz*phase.x - tmp.yw*phase.y);   // not completely correct, but seems to increase sharpness a bit
  vec2 xy = tmp.xz*phase.y + tmp.yw*phase.x;

  vec3 du = vec3(1.0 + rxy.x, 0.0, xy.x);
  vec3 dv = vec3(0.0, 1.0 + rxy.y, xy.y);
  vec3 m = cross(du, dv);
  return normalize(vec3(waveDirection.x*m.x + waveDirection.y*m.y,
                       -waveDirection.y*m.x + waveDirection.x*m.y,
                        m.z));
}
