// uint3 : SV_GroupThreadID     Thread id within block
// uint3 : SV_GroupID           Block id
// uint3 : SV_DispatchThreadID  Global thread id
// uint  : SV_GroupIndex        Linearized thread id within block

// groupshared shmem[64];  Declaration of shared memory

//AllMemoryBarrier - block execution until all memory accesses has completed. (my writes visible to world?)
//AllMemoryBarrierWithGroupSync - Above + wait until threads in group have reached this call.
//DeviceMemoryBarrier
//DeviceMemoryBarrierWithGroupSync
//GroupMemoryBarrier
//GroupMemoryBarrierWithGroupSync();

Texture2D HDR;

RWStructuredBuffer<float>  scratchBuffer;

groupshared float reduce[256];

[numthreads(256,1,1)]
void main(uint3 gid : SV_GroupID, uint3 tid : SV_GroupThreadID, uint tix : SV_GroupIndex)
{
  // 16x16=256 threads.
  uint tx = tid.x & 0xfu;
  uint ty = (tid.x >> 4);

  // Block of image that this group shold process
  uint width, height;
  HDR.GetDimensions(width, height);
  uint X0 = (width*gid.x) / 16;
  uint X1 = (width*(gid.x + 1)) / 16;
  uint Y0 = (height*gid.y) / 16;
  uint Y1 = (height*(gid.y + 1)) / 16;

  float w = 1.0 / ((float)width*(float)height);

  {
    float avg = 0.0;
    for (uint j = Y0 + ty; j < Y1; j += 16) {
      for (uint i = X0 + tx; i < X1; i += 16) {
        float3 col = HDR.Load(int3(i, j, 0)).rgb;
        float luminance = dot(float3(0.299, 0.587, 0.114), col);
        avg += w*log(max(1e-7, luminance));
      }
    }
    reduce[tid.x] = avg;
  }

  GroupMemoryBarrierWithGroupSync();
  if (ty == 0) {
    float avg = 0.0;
    for (int i = 0; i < 16; i++) {
      avg += reduce[16*i + tx]; // might have 2-way bank conflict. 
    }
    reduce[tx] = avg; // only tx reads and writes to that shmem element, no need for sync.
  }

  GroupMemoryBarrierWithGroupSync();
  if ((ty == 0) && (tx == 0)) {
    float avg = 0.0;
    for (int i = 0; i < 16; i++) {
      avg += reduce[i];
    }
    scratchBuffer[16 * gid.y + gid.x] = avg;
  }
}
