
#include "SpriteCommon.hlsl"




void geometryFunction(point VertexIn In[1], inout TriangleStream<VertexOut> OutputStream)
{
  if (In[0].Position.z < -In[0].Position.w) return;

#if 0

  for (int i = 0; i < 4; i++) {
    VertexOut vertex = (VertexOut)0;

    transferAttributes(In[i], vertex);
    quadExpand(vertex.Position,
               vertex.Weights,
               vertex.TexCoord.xy,
               In[0].Position,
               In[0].Parameters,
               In[0].TexCoord,
               (float)i);
    OutputStream.Append(vertex);
  }

#else
  float3 p = In[0].Position.xyz / In[0].Position.w;
  float2 size = In[0].Parameters.xy;
  float2 offset = In[0].Parameters.zw;

  float2 viewportScale = 1.0 / (0.5 * viewportSize);
  float2 scaledSize = size * viewportScale;

  float2 center = viewportSize * (0.5 + 0.5 * p.xy);
  float aspect = viewportSize.x / viewportSize.y;

#ifdef COGS_SPRITES_NOALIGNPIXEL
  float2 ll = center + offset + float2(-0.5, -0.5) * size;
#else
  float2 ll = floor(center + offset + float2(-0.5, -0.5) * size);
#endif

  float depth = alwaysOnTop ? (1.0-COGS_CLEAR_DEPTH) : p.z;

#ifdef COGS_SPRITES_IMAGETRANSFORM
  float2 scaledCenter = (center + offset) * viewportScale;
  float2 v0 = scaledCenter + TransformImage(float2(-0.5, -0.5) * scaledSize, aspect) - float2(1, 1);
  float2 v1 = scaledCenter + TransformImage(float2(0.5, -0.5) * scaledSize, aspect) - float2(1, 1);
  float2 v2 = scaledCenter + TransformImage(float2(-0.5, 0.5) * scaledSize, aspect) - float2(1, 1);
  float2 v3 = scaledCenter + TransformImage(float2(0.5, 0.5) * scaledSize, aspect) - float2(1, 1);
#else
  float2 v0 = ll * viewportScale - float2(1, 1);
  float2 v1 = (ll + float2(1, 0) * size) * viewportScale - float2(1, 1);
  float2 v2 = (ll + float2(0, 1) * size) * viewportScale - float2(1, 1);
  float2 v3 = (ll + float2(1, 1) * size) * viewportScale - float2(1, 1);
#endif

  VertexOut vertex = (VertexOut)0;
  vertex.Position = float4(v0, depth, 1.0);

  vertex.Weights = float4(0, 0, 0, 0);
  transferAttributes(In[0], vertex);
  vertex.TexCoord.xy = In[0].TexCoord.xy;
  OutputStream.Append(vertex);

  vertex.Position = float4(v1, depth, 1.0);
  vertex.Weights = float4(1, 0, 0, 0);
  transferAttributes(In[0], vertex);
  vertex.TexCoord.xy = In[0].TexCoord.zy;
  OutputStream.Append(vertex);

  vertex.Position = float4(v2, depth, 1.0);
  vertex.Weights = float4(0, 1, 0, 0);
  transferAttributes(In[0], vertex);
  vertex.TexCoord.xy = In[0].TexCoord.xw;
  OutputStream.Append(vertex);

  vertex.Position = float4(v3, depth, 1.0);
  vertex.Weights = float4(1, 1, 0, 0);
  transferAttributes(In[0], vertex);
  vertex.TexCoord.xy = In[0].TexCoord.zw;
  OutputStream.Append(vertex);
#endif
}
