#define COGS_COMBINE(a, b) a##b

float3 GetNormal(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_NORMAL_MAP
  float3 normal = normalize(In.Normal);
  float3 normalColor = 2.0 * normalMap.Sample(normalMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_NORMAL_UV)).rgb - 1.0;

  float3 tangent = normalize(In.Tangent);

#ifdef COGS_STANDARD_MATERIAL_HAS_BITANGENT
  float3 bitangent = normalize(In.Bitangent);
#else
  float3 bitangent = normalize(cross(normal, tangent));
#endif

  float3 bumpNormal = normalize(normalColor.z * normal + normalColor.x * tangent + normalColor.y * bitangent);
  normal = normalize(lerp(normal, bumpNormal, normalMapFactor));
  
  return In.IsFrontFace ? normal : -normal;
#else
  return normalize(In.IsFrontFace ? In.Normal : -In.Normal);
#endif
}

float4 GetAlbedo(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_ALBEDO_MAP
  return albedo * albedoMap.Sample(albedoMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_ALBEDO_UV).xy);
#else
  return albedo;
#endif
}

float GetRoughness(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_ROUGHNESS_MAP
  return roughness * roughnessMap.Sample(roughnessMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_ROUGHNESS_UV)).r;
#elif defined(COGS_STANDARD_MATERIAL_METALLIC_ROUGHNESS_MAP)
  return roughness * metallicRoughnessMap.Sample(metallicRoughnessMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_METALLIC_ROUGHNESS_UV)).g;
#else
  return roughness;
#endif
}

float GetMetallic(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_METALLIC_MAP
  return metallic * metallicMap.Sample(metallicMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_METALLIC_UV)).r;
#elif defined(COGS_STANDARD_MATERIAL_METALLIC_ROUGHNESS_MAP)
  return metallic * metallicRoughnessMap.Sample(metallicRoughnessMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_METALLIC_ROUGHNESS_UV)).b;
#else
  return metallic;
#endif
}

float GetGloss(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_GLOSS_MAP
  return gloss * glossMap.Sample(glossMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_GLOSS_UV)).r;
#elif defined(COGS_STANDARD_MATERIAL_SPECULAR_GLOSS_MAP)
  return gloss * specularGlossMap.Sample(specularGlossMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_SPECULAR_GLOSS_UV)).a;
#else
  return gloss;
#endif
}

float3 GetSpecular(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_SPECULAR_MAP
  return specular * specularMap.Sample(specularMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_SPECULAR_UV)).rgb;
#elif defined(COGS_STANDARD_MATERIAL_SPECULAR_GLOSS_MAP)
  return specular * specularGlossMap.Sample(specularGlossMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_SPECULAR_GLOSS_UV)).rgb;
#else
  return specular;
#endif
}

float3 GetEmissive(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_EMISSIVE_MAP
  return emissive * emissiveMap.Sample(emissiveMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_EMISSIVE_UV).xy).rgb;
#else
  return emissive;
#endif
}

float GetOcclusion(in VertexIn In)
{
#ifdef COGS_STANDARD_MATERIAL_OCCLUSION_MAP
  return occlusion * occlusionMap.Sample(occlusionMapSampler, COGS_COMBINE(In.TexCoord, COGS_STANDARD_MATERIAL_OCCLUSION_UV)).r;
#else
  return occlusion;
#endif
}
